(*
 *
 *)

(* 
   3次元実射影空間
   [| x0; x1; x2; x3 |] (x3 <> 0)
*)

module RP3 : sig
  type t = private float array
  val of_array : float array -> t
  val to_array : t -> float array
  val to_tuple : t -> float * float * float
  val motion_x : float -> t -> t
  val rot_x : float -> t -> t
  val motion_y : float -> t -> t
  val rot_y : float -> t -> t
  val motion_z : float -> t -> t
  val rot_z : float -> t -> t
end
= struct
  type t = float array
  let dim = 4
  let of_array v = assert (Array.length v = dim && v.(3) <> 0.); [| v.(0) /. v.(3); v.(1) /. v.(3); v.(2) /. v.(3); 1. |]
  let to_array v = [| v.(0) /. v.(3); v.(1) /. v.(3); v.(2) /. v.(3); 1. |]
  let to_tuple v = (v.(0) /. v.(3), v.(1) /. v.(3), v.(2) /. v.(3))
  let inner_prod u v k =
    let uv = Array.init 4 (fun n -> u.(k).(n) *. v.(n)) in
    Array.fold_left (+.) 0. uv

  let prod m v = Array.init 4 (fun k -> inner_prod m v k)

  let a1 x = [|
    [|1.;0.;0.;x|];
    [|0.;1.;0.;0.|];
    [|0.;0.;1.;0.|];
    [|0.;0.;0.;1.|]
  |]

  let a2 y = [|
    [|1.;0.;0.;0.|];
    [|0.;1.;0.;y|];
    [|0.;0.;1.;0.|];
    [|0.;0.;0.;1.|]
  |]

  let a3 z = [|
    [|1.;0.;0.;0.|];
    [|0.;1.;0.;0.|];
    [|0.;0.;1.;z|];
    [|0.;0.;0.;1.|]
  |]

  let para (x,y,z) = [|
    [|1.;0.;0.;x|];
    [|0.;1.;0.;y|];
    [|0.;0.;1.;z|];
    [|0.;0.;0.;1.|]
  |]

  let a4 phi = [|
    [|1.;0.;0.;0.|];
    [|0.;cos phi; -.sin phi;0.|];
    [|0.;sin phi; cos phi;  0.|];
    [|0.;0.;0.;1.|]
  |]

  let a5 psi = [|
    [|cos psi; 0.; sin psi; 0.|];
    [|0.; 1.; 0.; 0.|];
    [|-.sin psi;0.; cos psi;  0.|];
    [|0.;0.;0.;1.|]
  |]

  let a6 theta = [|
    [|cos theta; -.sin theta; 0.; 0.|];
    [|sin theta; cos theta; 0.; 0.|];
    [|0.;0.;1.;0.|];
    [|0.;0.;0.;1.|]
  |]


  let motion_x x p =
    let p_arr = to_array p in
    of_array (prod (a1 x) p_arr)
  
  let rot_x phi p =
    let p_arr = to_array p in
    of_array (prod (a4 phi) p_arr)

  let motion_y y p =
    let p_arr = to_array p in
    of_array (prod (a2 y) p_arr)
 
  let rot_y psi p =
    let p_arr = to_array p in
    of_array (prod (a5 psi) p_arr)

  let motion_z z p = 
    let p_arr = to_array p in
    of_array (prod (a3 z) p_arr)

  let rot_z theta p =
    let p_arr = to_array p in
    of_array (prod (a6 theta) p_arr)
end


let init_gl width height =
  GlDraw.shade_model `smooth;
  GlClear.color (0.0, 0.0, 0.0);
  GlClear.depth 1.0;
  GlClear.clear [`color; `depth];
  Gl.enable `depth_test;
  GlFunc.depth_func `lequal;
  GlMisc.hint `perspective_correction `nicest

let z = 0.2
let pi = 4. *. atan 1.

let theta = ref 0.
let omega = 0.02 *. pi


let draw_gl_scene () =
  let rec draw_cylinder y0 y1 r0 r1 th0 n idx =
    let s = 2. *. pi *. (float idx) /. (float n) in
    let p0 = RP3.of_array [| r0; 0.; 0.; 1.|] |> RP3.rot_y (s +. th0) |> RP3.motion_y y0 in
    let p1 = RP3.of_array [| r1; 0.; 0.; 1.|] |> RP3.rot_y (s +. th0) |> RP3.motion_y y1 in
    if idx >= n then
      begin
	RP3.to_tuple p0 |> GlDraw.vertex3;
	RP3.to_tuple p1 |> GlDraw.vertex3;
	GlDraw.ends()
      end
    else
      begin
	if idx = 0 then
	  begin
	    GlDraw.begins `quad_strip;
	    GlDraw.color(0.7, 0.7, 0.7);
	  end
	else if idx = 2 then GlDraw.color (0.5, 0.5, 0.5);
	RP3.to_tuple p0 |> GlDraw.vertex3;
	RP3.to_tuple p1 |> GlDraw.vertex3;
	draw_cylinder y0 y1 r0 r1 th0 n (idx+1)
      end
  in

  (* *)
  let rect t =
    if t <= 0.5 *. pi then t
    else if 0.5 *. pi < t && t <= 1.5 *. pi then pi -. t
    else if pi < 2. *. pi then t -. 2. *. pi
    else 0.
  in
	
  let rec draw_cube (x0, y0, z0) lx ly lz =
    GlDraw.begins `quads;
    
    GlDraw.color (0.5,0.5,0.5);                
    GlDraw.vertex3 (x0+.lx, y0+.ly, z0-.lz);
    GlDraw.vertex3 (x0-.lx, y0+.ly, z0-.lz);
    GlDraw.vertex3 (x0-.lx, y0+.ly, z0-.lz);
    GlDraw.vertex3 (x0+.lx, y0+.ly, z0-.lz);  
    
    GlDraw.color (0.5,0.5,0.5);      
    GlDraw.vertex3 (x0+.lx, y0-.ly, z0+.lz);
    GlDraw.vertex3 (x0-.lx, y0-.ly, z0+.z);
    GlDraw.vertex3 (x0-.lx, y0-.ly, z0-.z);
    GlDraw.vertex3 (x0+.lx, y0-.ly, z0-.z);  
    
    GlDraw.color (0.5,0.5,0.5);
    GlDraw.vertex3 (x0+.lx, y0+.ly, z0+.lz);
    GlDraw.vertex3 (x0-.lx, y0+.ly, z0+.lz);
    GlDraw.vertex3 (x0-.lx, y0-.ly, z0+.lz);
    GlDraw.vertex3 (x0+.lx, y0-.ly, z0+.lz);
    
    GlDraw.color (0.5,0.5,0.5);
    GlDraw.vertex3 (x0+.lx, y0-.ly, z0-.lz); 
    GlDraw.vertex3 (x0-.lx, y0-.ly, z0-.lz); 
    GlDraw.vertex3 (x0-.lx, y0+.ly, z0-.lz); 
    GlDraw.vertex3 (x0+.lx, y0+.ly, z0-.lz); 
    
    GlDraw.color (0.5,0.5,0.5);                   
    GlDraw.vertex3 (x0-.lx, y0+.ly, z0+.lz); 
    GlDraw.vertex3 (x0-.lx, y0+.ly, z0-.lz); 
    GlDraw.vertex3 (x0-.lx, y0-.ly, z0-.lz); 
    GlDraw.vertex3 (x0-.lx, y0-.ly, z0+.lz); 
    
    GlDraw.color (0.5,0.5,0.5);         
    GlDraw.vertex3 (x0+.lx, y0+.ly, z0-.lz); 
    GlDraw.vertex3 (x0+.lx, y0+.ly, z0+.lz); 
    GlDraw.vertex3 (x0+.lx, y0-.ly, z0+.lz); 
    GlDraw.vertex3 (x0+.lx, y0-.ly, z0-.lz); 
    GlDraw.ends ()
  in

  (* y 方向の移動 *)
  let dy = -1. in
	
  GlClear.clear [`color; `depth];
  
  (* Draw cylinder *)
  GlMat.load_identity ();
  GlMat.translate3 (0., 0., -7.0);
  GlMat.rotate3 45.0 (0., 1., 0.);
  draw_cube (2.,4.,-2.) 1. 1. 1.;
  let rate = 1. in
  draw_cylinder (4.+.dy) (1.+.dy) 0.5 0.5 (rate *. !theta) 30 0;
  draw_cylinder (1.+.dy) (0.5+.dy) 0.3 0.2 (rate *. !theta) 30 0; 
  
  draw_cube (rect !theta +. 0.5 ,-.1., -1.) 1. 0.2 1.;
  Glut.swapBuffers ();
  theta := if !theta < 2. *. pi then !theta +. omega else 0.


(* Handle window reshape events *)
let reshape_cb ~w ~h =
  let ratio = (float_of_int w) /. (float_of_int h) in
  GlDraw.viewport 0 0 w h;
  GlMat.mode `projection;
  GlMat.load_identity ();
  GluMat.perspective 45.0 ratio (0.1, 100.0);
  GlMat.mode `modelview;
  GlMat.load_identity ()

(* Handle keyboard events *)
let keyboard_cb ~key ~x ~y =
  match key with
  | 27 (* ESC *) -> exit 0
  | _ -> ()

(* Draw the scene whever idle *)
let idle_cb () =
  draw_gl_scene ()

let () =
  let width = 640 and height = 480 in
  ignore (Glut.init Sys.argv);
  Glut.initDisplayMode ~alpha:true ~depth:true ~double_buffer:true ();
  Glut.initWindowSize width height;
  ignore (Glut.createWindow "LablGL Machine Tool Simulation");
  Glut.displayFunc draw_gl_scene;
  Glut.keyboardFunc keyboard_cb;
  Glut.reshapeFunc reshape_cb;
  Glut.idleFunc(Some idle_cb);
  init_gl width height;
  Glut.mainLoop ()

